@@ -17,6 +17,7 @@
 #include <string.h>
 #include <ctype.h>
 #include <errno.h>
+#include <limits.h>
 #include <sys/ioctl.h>
 #include <sys/fcntl.h>
 #include <sys/file.h>
@@ -39,7 +40,8 @@
 
 /*--------------------------------------------------------------------------*/
 
-static int 			line_number;			/* config file line number		*/
+static int 			current_line_number; /* current line number				*/
+static int 			section_line_number; /* line number at start of section	*/
 static BOOTRECORD	boot_record;
 
 /*--------------------------------------------------------------------------*/
@@ -68,7 +70,8 @@ enum keyword_id {
 	KW_READ_ONLY,
 	KW_READ_WRITE,
 	KW_ROOT,
-	KW_CONSOLE
+	KW_CONSOLE,
+	KW_OPTIONAL
 };
 
 static struct keyword {
@@ -99,7 +102,8 @@ static struct keyword {
 	{"read-only",		KW_READ_ONLY		},
 	{"read-write",		KW_READ_WRITE		},
 	{"root",			KW_ROOT				},
-	{"console",			KW_CONSOLE			}
+	{"console",			KW_CONSOLE			},
+	{"optional",		KW_OPTIONAL			}
 };
 
 #define NUMKEYWORDS	(sizeof(keywords) / sizeof(keywords[0]))
@@ -157,7 +161,7 @@ is_reserved_label
 static
 char *
 strdupx
-(	char	*str
+(	const char	*str
 )
 {	char	*p;
 
@@ -301,11 +305,22 @@ despace
 
 static
 void
-conferror
-(	const char		*s
+section_error
+(	const char	*s
 )
 {
-	die("%s:%d: %s\n", config_file, line_number, s);
+	die("%s:%d: %s\n", config_file, section_line_number, s);
+}
+
+/*--------------------------------------------------------------------------*/
+
+static
+void
+line_error
+(	const char	*s
+)
+{
+	die("%s:%d: %s\n", config_file, current_line_number, s);
 }
 
 /*--------------------------------------------------------------------------*/
@@ -316,7 +331,7 @@ error_spurious_data
 (	void
 )
 {
-	conferror("spurious data at end of line\n");
+	line_error("spurious data at end of line\n");
 }
 
 /*--------------------------------------------------------------------------*/
@@ -338,7 +353,8 @@ redefined_option
 (	const char	*name
 )
 {
-	printf("%s:%d: Redefinition of `%s'\n", config_file, line_number, name);
+	message("%s:%d: Redefinition of `%s'\n",
+			 config_file, current_line_number, name);
 }
 
 /*--------------------------------------------------------------------------*/
@@ -350,7 +366,7 @@ reserved_label
 )
 {
 	die("%s:%d: Boot record label or alias with reserved name `%s'\n",
-			 config_file, line_number, name);
+			 config_file, current_line_number, name);
 }
 
 /*--------------------------------------------------------------------------*/
@@ -390,7 +406,7 @@ get_num
 
 	if (*p == '\0')
 	{
-		conferror("missing numeric constant\n");
+		line_error("missing numeric constant\n");
 	}
 
 	/* determine number base from prefix */
@@ -476,7 +492,7 @@ get_num
 	if (digits == 0)
 	{
 fail:
-		conferror("invalid numeric constant\n");
+		line_error("invalid numeric constant\n");
 	}
 
 	*str = skip_white(p);
@@ -509,7 +525,7 @@ get_bool
 	}
 
 	die("Boolean argument expected on line %d of %s\n",
-		line_number, config_file);
+		current_line_number, config_file);
 }
 
 /*--------------------------------------------------------------------------*/
@@ -547,34 +563,106 @@ add_boot_record
 }
 
 /*--------------------------------------------------------------------------*/
-/* Add a New File Definition
- */
 
 static
-void
-add_file_def
+FILEDEF	**
+find_file_def
 (	const char	*path
 )
 {	FILEDEF		**p;
-	FILEDEF		*new;
 
 	for (p = &config.files; *p; p = &(*p)->next)
 	{
 		if (case_equal_strings((*p)->path, path))
 		{
-			return;
+			break;
 		}
 	}
 
-	if ((new = malloc(sizeof(FILEDEF))) == NULL)
+	return p;
+}
+
+/*--------------------------------------------------------------------------*/
+/* Add a New File Definition
+ */
+
+static
+const char *
+add_file_def
+(	const char	*path
+)
+{	FILEDEF		*file;
+	FILEDEF		**link;
+	char		*rpath;
+
+/* The following section has been removed for now because
+ * it breaks the --root option. realpath() reads the file system
+ * to resolve symblic links but because we do not apply the root_path
+ * configuration until later we can end up resolving the wrong symbolic
+ * link.
+ */
+#if 0
+	if ((rpath = malloc(PATH_MAX)) != NULL)
 	{
-		error_nomemory();
+		if (realpath(path, rpath) == NULL)
+		{
+			free(rpath);
+			rpath = NULL;
+		}
 	}
 
-	new->next = NULL;
-	new->path = path;
-	new->map  = NULL;
-	*p		  = new;
+	if (rpath == NULL)
+#endif
+	{
+		rpath = strdupx(path);
+	}
+
+	link = find_file_def(rpath);
+
+	if (*link == NULL)
+	{
+		if ((file = malloc(sizeof(FILEDEF))) == NULL)
+		{
+			error_nomemory();
+		}
+	
+		file->path      = rpath;
+		file->map       = NULL;
+		file->refcount  = 1;
+		file->next      = NULL;
+		*link           = file;
+	}
+	else
+	{
+		(*link)->refcount++;
+	}
+
+	return rpath;
+}
+
+/*--------------------------------------------------------------------------*/
+/* Dereference an existing file definition.
+ */
+
+void
+remove_file_def
+(	const char	*path
+)
+{	FILEDEF		*file;
+	FILEDEF		**link;
+
+	if (path && *path)
+	{
+		link = find_file_def(path);
+	
+		if ((file = *link) != NULL)
+		{
+			if (file->refcount)
+			{
+				file->refcount--;
+			}
+		}
+	}
 }
 
 /*--------------------------------------------------------------------------*/
@@ -582,10 +670,11 @@ add_file_def
 static
 int
 parse_options
-(	char	*name,
-	char	*text
+(	char			*name,
+	char			*text
 )
-{
+{	unsigned long	flag;
+
 	switch (keyword(name))
 	{
 		case KW_DEFAULT:
@@ -601,7 +690,7 @@ parse_options
 			{
 				redefined_option(name);
 			}
-			add_file_def(config.options.boot_message = strdupx(text));
+			config.options.boot_message = add_file_def(text);
 			break;
 
 		case KW_BOOT:
@@ -613,19 +702,21 @@ parse_options
 			break;
 
 		case KW_DEBUG:
-			if (config.options.boot_debug)
+			flag = *text ? get_bool(text) : TRUE;
+			if (config.options.boot_debug && (*config.options.boot_debug != flag))
 			{
 				redefined_option(name);
 			}
-			config.options.boot_debug = copy_long(*text ? get_bool(text) : TRUE);
+			config.options.boot_debug = copy_long(flag);
 			break;
 
 		case KW_PROMPT:
-			if (config.options.boot_prompt)
+			flag = *text ? get_bool(text) : TRUE;
+			if (config.options.boot_prompt && (*config.options.boot_prompt != flag))
 			{
 				redefined_option(name);
 			}
-			config.options.boot_prompt = copy_long(*text ? get_bool(text) : TRUE);
+			config.options.boot_prompt = copy_long(flag);
 			break;
 
 		case KW_DELAY:
@@ -662,7 +753,7 @@ parse_options
 
 			if (strlen(text) >= CL_SIZE)
 			{
-				conferror("Boot argument is too long");
+				line_error("Boot argument is too long");
 			}
 			config.default_cmdline = strdupx(text);
 			break;
@@ -675,7 +766,7 @@ parse_options
 
 			if (strlen(text) >= CL_SIZE)
 			{
-				conferror("Boot argument is too long");
+				line_error("Boot argument is too long");
 			}
 			config.default_append = strdupx(text);
 			break;
@@ -697,11 +788,12 @@ parse_options
 			break;
 
 		case KW_RESTRICTED:
-			if (config.default_restricted)
+			flag = *text ? get_bool(text) : TRUE;
+			if (config.default_restricted && (*config.default_restricted != flag))
 			{
 				redefined_option(name);
 			}
-			config.default_restricted = copy_long(*text ? get_bool(text) : TRUE);
+			config.default_restricted = copy_long(flag);
 			break;
 
 		case KW_KERNEL:
@@ -720,7 +812,7 @@ parse_options
 
 			if (strlen(text) >= CL_SIZE)
 			{
-				conferror("Root argument is too long");
+				line_error("Root argument is too long");
 			}
 			config.default_root = strdupx(text);
 			break;
@@ -738,27 +830,27 @@ parse_options
 
 			if (strlen(text) >= CL_SIZE)
 			{
-				conferror("Console argument is too long");
+				line_error("Console argument is too long");
 			}
 			config.default_console = strdupx(text);
 			break;
 
 		case KW_READ_WRITE:
-			if (config.default_read_only)
+			flag = *text ? !get_bool(text) : FALSE;
+			if (config.default_read_only && (*config.default_read_only != flag))
 			{
 				redefined_option(name);
 			}
-
-			config.default_read_only = copy_long(*text ? !get_bool(text) : FALSE);
+			config.default_read_only = copy_long(flag);
 			break;
 
 		case KW_READ_ONLY:
-			if (config.default_read_only)
+			flag = *text ? get_bool(text) : TRUE;
+			if (config.default_read_only && (*config.default_read_only != flag))
 			{
 				redefined_option(name);
 			}
-
-			config.default_read_only = copy_long(*text ? get_bool(text) : TRUE);
+			config.default_read_only = copy_long(flag);
 			break;
 
 		default:
@@ -773,12 +865,22 @@ parse_options
 static
 int
 parse_boot
-(	char	*name,
-	char	*text
+(	char			*name,
+	char			*text
 )
-{
+{	unsigned long	flag;
+
 	switch (keyword(name))
 	{
+		case KW_OPTIONAL:
+			flag = *text ? get_bool(text) : TRUE;
+			if (boot_record.optional && (*boot_record.optional != flag))
+			{
+				redefined_option(name);
+			}
+			boot_record.optional = copy_long(flag);
+			break;
+
 		case KW_LABEL:
 			if (boot_record.label)
 			{
@@ -812,12 +914,12 @@ parse_boot
 			break;
 
 		case KW_RESTRICTED:
-			if (boot_record.restricted)
+			flag = *text ? get_bool(text) : TRUE;
+			if (boot_record.restricted && (*boot_record.restricted != flag))
 			{
 				redefined_option(name);
 			}
-
-			boot_record.restricted = copy_long(*text ? get_bool(text) : TRUE);
+			boot_record.restricted = copy_long(flag);
 			break;
 
 		case KW_CMDLINE:
@@ -828,7 +930,7 @@ parse_boot
 
 			if (strlen(text) >= CL_SIZE)
 			{
-				conferror("Boot argument is too long");
+				line_error("Boot argument is too long");
 			}
 			boot_record.cmdline = strdupx(text);
 			break;
@@ -841,7 +943,7 @@ parse_boot
 
 			if (strlen(text) >= CL_SIZE)
 			{
-				conferror("Boot argument is too long");
+				line_error("Boot argument is too long");
 			}
 			boot_record.append = strdupx(text);
 			break;
@@ -864,7 +966,7 @@ parse_boot
 			{
 				redefined_option(name);
 			}
-			add_file_def(boot_record.ramdisk = strdupx(text));
+			boot_record.ramdisk = add_file_def(text);
 			break;
 
 		case KW_SYMBOLS:
@@ -872,7 +974,7 @@ parse_boot
 			{
 				redefined_option(name);
 			}
-			add_file_def(boot_record.symtab = strdupx(text));
+			boot_record.symtab = add_file_def(text);
 			break;
 
 		case KW_KERNEL:
@@ -880,15 +982,16 @@ parse_boot
 			{
 				redefined_option(name);
 			}
-			add_file_def(boot_record.kernel = strdupx(text));
+			boot_record.kernel = add_file_def(text);
 			break;
 
 		case KW_CALLDBG:
-			if (boot_record.callmonitor)
+			flag = *text ? get_bool(text) : TRUE;
+			if (boot_record.callmonitor && (*boot_record.callmonitor != flag))
 			{
 				redefined_option(name);
 			}
-			boot_record.callmonitor = copy_long(*text ? get_bool(text) : TRUE);
+			boot_record.callmonitor = copy_long(flag);
 			break;
 
 		case KW_ROOT:
@@ -899,7 +1002,7 @@ parse_boot
 
 			if (strlen(text) >= CL_SIZE)
 			{
-				conferror("Root argument is too long");
+				line_error("Root argument is too long");
 			}
 			boot_record.root = strdupx(text);
 			break;
@@ -917,27 +1020,27 @@ parse_boot
 
 			if (strlen(text) >= CL_SIZE)
 			{
-				conferror("Console argument is too long");
+				line_error("Console argument is too long");
 			}
 			boot_record.console = strdupx(text);
 			break;
 
 		case KW_READ_WRITE:
-			if (boot_record.read_only)
+			flag = *text ? !get_bool(text) : FALSE;
+			if (boot_record.read_only && (*boot_record.read_only != flag))
 			{
 				redefined_option(name);
 			}
-
-			boot_record.read_only = copy_long(*text ? !get_bool(text) : FALSE);
+			boot_record.read_only = copy_long(flag);
 			break;
 
 		case KW_READ_ONLY:
-			if (boot_record.read_only)
+			flag = *text ? get_bool(text) : TRUE;
+			if (boot_record.read_only && (*boot_record.read_only != flag))
 			{
 				redefined_option(name);
 			}
-
-			boot_record.read_only = copy_long(*text ? get_bool(text) : TRUE);
+			boot_record.read_only = copy_long(flag);
 			break;
 
 		default:
@@ -957,7 +1060,7 @@ boot_wrapup
 {
 	if (boot_record.label == NULL)
 	{
-		conferror("No label specified");
+		section_error("No label specified for boot record.");
 	}
 
 	/* add file definition for default kernel */
@@ -965,13 +1068,13 @@ boot_wrapup
 	{
 		if (config.default_kernel)
 		{
-			add_file_def(boot_record.kernel = config.default_kernel);
+			boot_record.kernel = add_file_def(config.default_kernel);
 		}
 	}
 
 	if (!add_boot_record(&boot_record))
 	{
-		conferror("Duplicate boot record label or alias");
+		section_error("Duplicate boot record label or alias.");
 	}
 
 	clear_boot_record();
@@ -1030,6 +1133,9 @@ parse_line
 		/* wrap up previous section */
 		wrapup_section();
 
+		/* remember where section starts */
+		section_line_number = current_line_number;
+
 		/* get section name */
 		extract(name, text + 1, sizeof(name), ']');
 
@@ -1044,7 +1150,7 @@ parse_line
 		}
 
 		die("unrecognised section `%s' in %s line %d\n",
-					name, config_file, line_number);
+					name, config_file, current_line_number);
 	}
 
 	/* get options name */
@@ -1053,7 +1159,7 @@ parse_line
 	if (!(*section_parse->parse)(name, text))
 	{
 		die("unrecognised option `%s' in %s line %d\n",
-					name, config_file, line_number);
+					name, config_file, current_line_number);
 	}
 }
 
@@ -1088,7 +1194,7 @@ read_config_file
 
 	if (f_verbose)
 	{
-		printf("Reading configuration file `%s'\n", config_file);
+		message("Reading configuration file `%s'\n", config_file);
 	}
 
 	/* allocate line buffers */
@@ -1100,7 +1206,8 @@ read_config_file
 	} 
 
 	/* reset current configuration file line number */
-	line_number = 0;
+	current_line_number = 0;
+	section_line_number = 1;
 
 	/* read line of input until end of file */
 	eof = FALSE;
@@ -1116,7 +1223,7 @@ read_config_file
 			}
 
 			/* update current line number */
-			line_number++;
+			current_line_number++;
 
 			/* check for long lines */
 			if (strlen(inptbf) == (MAXLINELEN - 1))
@@ -1141,7 +1248,7 @@ read_config_file
 			{
 			too_long:
 				die("line %d too long in file \"%s\"\n",
-						line_number, config_file);
+						current_line_number, config_file);
 			}
 
 			/* if appending to existing line */
