@@ -9,6 +9,12 @@
 
 /*--------------------------------------------------------------------------*/
 
+#ifndef USE_TRAP
+#define USE_TRAP			1
+#endif
+
+/*--------------------------------------------------------------------------*/
+
 #define BVMBugEntry			(0xF8000404)
 
 #define BB_PUTCHAR			0
@@ -34,15 +40,22 @@
 #define BB_DEFINE_SYM		20
 #define BB_DELETE_SYM		21
 #define BB_ENTER			22
+#define BB_TFTP_READ		23
+#define BB_NETBOOTINFO		24
+#define BB_BOOTARGS			25
+#define BB_BRD_ID			26
 
 /*--------------------------------------------------------------------------*/
 /* BVMBug error codes
  */
 
-#define DISK_READ_ERROR		1
-#define OUT_OF_MEMORY		2
-#define INVALID_FILE_MAP	3
-#define ILLEGAL_SYMBOL_NAME	4
+#define DISK_READ_ERROR		1		/* Disk Read Error						*/
+#define OUT_OF_MEMORY		2		/* Out of memory						*/
+#define INVALID_FILE_MAP	3		/* File map array contains invalid data	*/
+#define ILLEGAL_SYMBOL_NAME	4		/* Illegal symbol name					*/
+#define FILE_NOT_FOUND		5		/* Illegal or missing file				*/
+#define NET_READ_ERROR		6		/* Network Read Error					*/
+#define USER_ABORTED		7		/* Aborted by user						*/
 
 /*--------------------------------------------------------------------------*/
 
@@ -85,6 +98,18 @@ typedef struct {
 
 } BOOTBLOCK;
 
+/*--------------------------------------------------------------------------*/
+
+typedef struct netbootinfo {
+	unsigned long	cipa;				/* client IP address				*/
+	unsigned long	sipa;				/* server IP address				*/
+	unsigned long	gipa;				/* gateway IP address				*/
+	unsigned long	subnetmask;			/* subnet IP address mask			*/
+	unsigned long	reserved;			/* reserved for broadcast IP		*/
+	char			bootfile[128];		/* bootfile name					*/
+
+} NETBOOTINFO; 
+
 /*--------------------------------------------------------------------------*/
 /* Call BVMBug putchar subroutine.
  */
@@ -98,11 +123,19 @@ BVMBug_putchar
 {	register unsigned long	d0 asm("d0") = (unsigned long) BB_PUTCHAR;
 	register unsigned long	d1 asm("d1") = (unsigned long) c;
 
+#if USE_TRAP
+	asm volatile (	"trap #15"
+					: /* no outputs */
+					: "r" (d0), "r" (d1)
+					: "d0", "d1"
+	);
+#else
 	asm volatile (	"jsr (%0)"
 					: /* no outputs */
 					: "a" (BVMBugEntry), "r" (d0), "r" (d1)
 					: "d0", "d1"
 	);
+#endif
 }
 
 /*--------------------------------------------------------------------------*/
@@ -118,11 +151,19 @@ BVMBug_putstr
 {	register unsigned long	 d0 asm("d0") = (unsigned long) BB_PUTSTR;
 	register void			*a0 asm("a0") = (void *) str;
 
+#if USE_TRAP
+	asm volatile (	"trap #15"
+					: /* no outputs */
+					: "r" (d0), "r" (a0)
+					: "d0", "a0"
+	);
+#else
 	asm volatile (	"jsr (%0)"
 					: /* no outputs */
 					: "a" (BVMBugEntry), "r" (d0), "r" (a0)
 					: "d0", "a0"
 	);
+#endif
 }
 
 /*--------------------------------------------------------------------------*/
@@ -140,11 +181,20 @@ BVMBug_print
 	register void			*a0 asm("a0") = (void *) str;
 	register void			*a1 asm("a1") = (void *) vap;
 
+#if USE_TRAP
+	asm volatile (	"trap #15"
+					: "=r" (d0)
+					: "r" (d0), "r" (a0), "r" (a1)
+					: "d0", "a0", "a1"
+	);
+
+#else
 	asm volatile (	"jsr (%1)"
 					: "=r" (d0)
 					: "a" (BVMBugEntry), "r" (d0), "r" (a0), "r" (a1)
 					: "d0", "a0", "a1"
 	);
+#endif
 
 	return d0;
 }
@@ -166,12 +216,20 @@ BVMBug_format
 	register void			*a1 asm("a1") = (void *) vap;
 	register void			*a2 asm("a2") = (void *) buf;
 
+#if USE_TRAP
+	asm volatile (	"trap #15"
+					: "=r" (d0)
+					: "r" (d0), "r" (a0), "r" (a1), "r" (a2)
+					: "d0", "a0", "a1", "a2", "memory"
+	);
+#else
 	asm volatile (	"jsr (%1)"
 					: "=r" (d0)
 					: "a" (BVMBugEntry), "r" (d0), "r" (a0),
 										 "r" (a1), "r" (a2)
 					: "d0", "a0", "a1", "a2", "memory"
 	);
+#endif
 
 	return d0;
 }
@@ -188,11 +246,19 @@ BVMBug_getchar
 )
 {	register unsigned long	d0 asm("d0") = (unsigned long) BB_GETCHAR;
 
+#if USE_TRAP
+	asm volatile (	"trap #15"
+					: "=r" (d0)
+					: "r" (d0)
+					: "d0"
+	);
+#else
 	asm volatile (	"jsr (%1)"
 					: "=r" (d0)
 					: "a" (BVMBugEntry), "r" (d0)
 					: "d0"
 	);
+#endif
 
 	return d0;
 }
@@ -209,11 +275,19 @@ BVMBug_getchar_nowait
 )
 {	register unsigned long	d0 asm("d0") = (unsigned long) BB_GETCHAR_NW;
 
+#if USE_TRAP
+	asm volatile (	"trap #15"
+					: "=r" (d0)
+					: "r" (d0)
+					: "d0"
+	);
+#else
 	asm volatile (	"jsr (%1)"
 					: "=r" (d0)
 					: "a" (BVMBugEntry), "r" (d0)
 					: "d0"
 	);
+#endif
 
 	return d0;
 }
@@ -233,11 +307,19 @@ BVMBug_getstr
 	register unsigned long	d1 asm("d1") = (unsigned long) maxlen;
 	register void*			a0 asm("a0") = (void *) buffer;
 
+#if USE_TRAP
+	asm volatile (	"trap #15"
+					: "=r" (d0)
+					: "r" (d0), "r" (d1), "r" (a0)
+					: "d0", "d1", "a0"
+	);
+#else
 	asm volatile (	"jsr (%1)"
 					: "=r" (d0)
 					: "a" (BVMBugEntry), "r" (d0), "r" (d1), "r" (a0)
 					: "d0", "d1", "a0"
 	);
+#endif
 
 	return d0;
 }
@@ -259,11 +341,19 @@ BVMBug_atoi
 	register unsigned long	d2 asm("d2") = (unsigned long) base;
 	register void *			a0 asm("a0") = (void *) string;
 
+#if USE_TRAP
+	asm volatile (	"trap #15"
+					: "=r" (d0), "=r" (d1)
+					: "r" (d0), "r" (d2), "r" (a0)
+					: "d0", "d1", "d2", "a0"
+	);
+#else
 	asm volatile (	"jsr (%2)"
 					: "=r" (d0), "=r" (d1)
 					: "a" (BVMBugEntry), "r" (d0), "r" (d2), "r" (a0)
 					: "d0", "d1", "d2", "a0"
 	);
+#endif
 
 	*result = d1;
 	return d0;
@@ -283,11 +373,19 @@ BVMBug_malloc
 	register unsigned long	d1 asm("d1") = (unsigned long) count;
 	register void *			a0 asm("a0");
 
+#if USE_TRAP
+	asm volatile (	"trap #15"
+					: "=r" (a0)
+					: "r" (d0), "r" (d1)
+					: "d0", "d1", "a0"
+	);
+#else
 	asm volatile (	"jsr (%1)"
 					: "=r" (a0)
 					: "a" (BVMBugEntry), "r" (d0), "r" (d1)
 					: "d0", "d1", "a0"
 	);
+#endif
 
 	return a0;
 }
@@ -305,11 +403,19 @@ BVMBug_free
 {	register unsigned long	d0 asm("d0") = (unsigned long) BB_FREE;
 	register void *			a0 asm("a0") = (void *) memptr;
 
+#if USE_TRAP
+	asm volatile (	"trap #15"
+					: /* no outputs */
+					: "r" (d0), "r" (a0)
+					: "d0", "a0"
+	);
+#else
 	asm volatile (	"jsr (%0)"
 					: /* no outputs */
 					: "a" (BVMBugEntry), "r" (d0), "r" (a0)
 					: "d0", "a0"
 	);
+#endif
 }
 
 /*--------------------------------------------------------------------------*/
@@ -329,11 +435,19 @@ BVMBug_memcmp
 	register void *			a0 asm("a0") = (void *) ptr1;
 	register void *			a1 asm("a1") = (void *) ptr2;
 
+#if USE_TRAP
+	asm volatile (	"trap #15"
+					: "=r" (d0)
+					: "r" (d0), "r" (d1), "r" (a0), "r" (a1)
+					: "d0", "d1", "a0", "a1", "memory"
+	);
+#else
 	asm volatile (	"jsr (%1)"
 					: "=r" (d0)
 					: "a" (BVMBugEntry), "r" (d0), "r" (d1), "r" (a0), "r" (a1)
 					: "d0", "d1", "a0", "a1", "memory"
 	);
+#endif
 
 	return d0;
 }
@@ -346,8 +460,8 @@ static
 __inline__
 void
 BVMBug_memset
-(	void	*dst,
-	int		val,
+(	void			*dst,
+	int				val,
 	unsigned long	siz
 )
 {	register unsigned long	d0 asm("d0") = (unsigned long) BB_MEMSET;
@@ -355,11 +469,19 @@ BVMBug_memset
 	register unsigned long	d2 asm("d2") = (unsigned long) siz;
 	register void *			a0 asm("a0") = (void *) dst;
 
+#if USE_TRAP
+	asm volatile (	"trap #15"
+					: /* no outputs */
+					: "r" (d0), "r" (d1), "r" (d2), "r" (a0)
+					: "d0", "d1", "d2", "a0", "memory"
+	);
+#else
 	asm volatile (	"jsr (%0)"
 					: /* no outputs */
 					: "a" (BVMBugEntry), "r" (d0), "r" (d1), "r" (d2), "r" (a0)
 					: "d0", "d1", "d2", "a0", "memory"
 	);
+#endif
 }
 
 /*--------------------------------------------------------------------------*/
@@ -370,20 +492,28 @@ static
 __inline__
 void
 BVMBug_memmove
-(	void		*dst,
-	const void	*src,
-	unsigned long		siz
+(	void			*dst,
+	const void		*src,
+	unsigned long	siz
 )
 {	register unsigned long	d0 asm("d0") = (unsigned long) BB_MEMMOVE;
 	register unsigned long	d1 asm("d1") = (unsigned long) siz;
 	register void *			a0 asm("a0") = (void *) dst;
 	register void *			a1 asm("a1") = (void *) src;
 
+#if USE_TRAP
+	asm volatile (	"trap #15"
+					: /* no outputs */
+					: "r" (d0), "r" (d1), "r" (a0), "r" (a1)
+					: "d0", "d1", "a0", "a1", "memory"
+	);
+#else
 	asm volatile (	"jsr (%0)"
 					: /* no outputs */
 					: "a" (BVMBugEntry), "r" (d0), "r" (d1), "r" (a0), "r" (a1)
 					: "d0", "d1", "a0", "a1", "memory"
 	);
+#endif
 }
 
 /*--------------------------------------------------------------------------*/
@@ -398,11 +528,19 @@ BVMBug_disk_open
 )
 {	register unsigned long	d0 asm("d0") = (unsigned long) BB_DISK_OPEN;
 
+#if USE_TRAP
+	asm volatile (	"trap #15"
+					: "=r" (d0)
+					: "r" (d0)
+					: "d0"
+	);
+#else
 	asm volatile (	"jsr (%1)"
 					: "=r" (d0)
 					: "a" (BVMBugEntry), "r" (d0)
 					: "d0"
 	);
+#endif
 
 	return d0;
 }
@@ -424,11 +562,19 @@ BVMBug_disk_read
 	register unsigned long	d2 asm("d2") = (unsigned long) block;
 	register void *			a0 asm("a0") = (void *) buffer;
 
+#if USE_TRAP
+	asm volatile (	"trap #15"
+					: "=r" (d0)
+					: "r" (d0), "r" (d1), "r" (d2), "r" (a0)
+					: "d0", "d1", "d2", "a0", "memory"
+	);
+#else
 	asm volatile (	"jsr (%1)"
 					: "=r" (d0)
 					: "a" (BVMBugEntry), "r" (d0), "r" (d1), "r" (d2), "r" (a0)
 					: "d0", "d1", "d2", "a0", "memory"
 	);
+#endif
 
 	return d0;
 }
@@ -445,11 +591,19 @@ BVMBug_disk_close
 )
 {	register unsigned long	d0 asm("d0") = (unsigned long) BB_DISK_CLOSE;
 
+#if USE_TRAP
+	asm volatile (	"trap #15"
+					: /* no outputs */
+					: "r" (d0)
+					: "d0"
+	);
+#else
 	asm volatile (	"jsr (%0)"
 					: /* no outputs */
 					: "a" (BVMBugEntry), "r" (d0)
 					: "d0"
 	);
+#endif
 }
 
 /*--------------------------------------------------------------------------*/
@@ -469,11 +623,19 @@ BVMBug_file_open
 	register unsigned long	d2 asm("d2") = (unsigned long) map_entries;
 	register void * 		a0 asm("a0") = (void *) map;
 
+#if USE_TRAP
+	asm volatile (	"trap #15"
+					: "=r" (d0)
+					: "r" (d0), "r" (d1), "r" (d2), "r" (a0)
+					: "d0", "d1", "d2", "a0"
+	);
+#else
 	asm volatile (	"jsr (%1)"
 					: "=r" (d0)
 					: "a" (BVMBugEntry), "r" (d0), "r" (d1), "r" (d2), "r" (a0)
 					: "d0", "d1", "d2", "a0"
 	);
+#endif
 
 	return d0;
 }
@@ -495,11 +657,19 @@ BVMBug_file_read
 	register unsigned long	d2 asm("d2") = (unsigned long) offset;
 	register void *			a0 asm("a0") = (void *) buffer;
 
+#if USE_TRAP
+	asm volatile (	"trap #15"
+					: "=r" (d0)
+					: "r" (d0), "r" (d1), "r" (d2), "r" (a0)
+					: "d0", "d1", "d2", "a0", "memory"
+	);
+#else
 	asm volatile (	"jsr (%1)"
 					: "=r" (d0)
 					: "a" (BVMBugEntry), "r" (d0), "r" (d1), "r" (d2), "r" (a0)
 					: "d0", "d1", "d2", "a0", "memory"
 	);
+#endif
 
 	return d0;
 }
@@ -516,11 +686,19 @@ BVMBug_file_close
 )
 {	register unsigned long	d0 asm("d0") = (unsigned long) BB_FILE_CLOSE;
 
+#if USE_TRAP
+	asm volatile (	"trap #15"
+					: /* no outputs */
+					: "r" (d0)
+					: "d0"
+	);
+#else
 	asm volatile (	"jsr (%0)"
 					: /* no outputs */
 					: "a" (BVMBugEntry), "r" (d0)
 					: "d0"
 	);
+#endif
 }
 
 /*--------------------------------------------------------------------------*/
@@ -537,11 +715,19 @@ BVMBug_gettime
 {	register unsigned long	d0 asm("d0") = (unsigned long) BB_GETTIME;
 	register unsigned long	d1 asm("d1");
 
+#if USE_TRAP
+	asm volatile (	"trap #15"
+					: "=r" (d0), "=r" (d1)
+					: "r" (d0)
+					: "d0", "d1"
+	);
+#else
 	asm volatile (	"jsr (%2)"
 					: "=r" (d0), "=r" (d1)
 					: "a" (BVMBugEntry), "r" (d0)
 					: "d0", "d1"
 	);
+#endif
 
 	if (time) *time = d0;
 	if (date) *date = d1;
@@ -564,11 +750,19 @@ BVMBug_define_symbol
 	register unsigned long	d2 asm("d2") = (unsigned long) type;
 	register void * 		a0 asm("a0") = (void *) name;
 
+#if USE_TRAP
+	asm volatile (	"trap #15"
+					: "=r" (d0)
+					: "r" (d0), "r" (d1), "r" (d2), "r" (a0)
+					: "d0", "d1", "d2", "a0"
+	);
+#else
 	asm volatile (	"jsr (%1)"
 					: "=r" (d0)
 					: "a" (BVMBugEntry), "r" (d0), "r" (d1), "r" (d2), "r" (a0)
 					: "d0", "d1", "d2", "a0"
 	);
+#endif
 
 	return d0;
 }
@@ -586,11 +780,19 @@ BVMBug_delete_symbol
 {	register unsigned long	d0 asm("d0") = (unsigned long) BB_DELETE_SYM;
 	register void * 		a0 asm("a0") = (void *) name;
 
+#if USE_TRAP
+	asm volatile (	"trap #15"
+					: "=r" (d0)
+					: "r" (d0), "r" (a0)
+					: "d0", "a0"
+	);
+#else
 	asm volatile (	"jsr (%1)"
 					: "=r" (d0)
 					: "a" (BVMBugEntry), "r" (d0), "r" (a0)
 					: "d0", "a0"
 	);
+#endif
 
 	return d0;
 }
@@ -607,12 +809,21 @@ BVMBug_enter
 )
 {	register unsigned long	d0 asm("d0") = (unsigned long) BB_ENTER;
 
+#if USE_TRAP
+	asm volatile (	"trap #15"
+					: /* no outputs */
+					: "r" (d0)
+					: "d0", "d1", "d2", "d3", "d4", "d5", "d6",
+					  "a0", "a1", "a2", "a3", "a4", "memory"
+	);
+#else
 	asm volatile (	"jsr (%0)"
 					: /* no outputs */
 					: "a" (BVMBugEntry), "r" (d0)
 					: "d0", "d1", "d2", "d3", "d4", "d5", "d6",
 					  "a0", "a1", "a2", "a3", "a4", "memory"
 	);
+#endif
 }
 
 /*--------------------------------------------------------------------------*/
@@ -648,4 +859,96 @@ BVMBug_reset
 		;
 }
 
+/*--------------------------------------------------------------------------*/
+
+static
+__inline__
+int
+BVMBug_tftp_read
+(	const char		*filename,
+	unsigned long	*count,
+	void			*buffer,
+	unsigned long	flags
+)
+{	register unsigned long	d0 asm("d0") = (unsigned long) BB_TFTP_READ;
+	register unsigned long	d1 asm("d1") = (unsigned long) *count;
+	register unsigned long	d2 asm("d2") = (unsigned long) flags;
+	register void * 		a0 asm("a0") = (void *) filename;
+	register void *			a1 asm("a1") = (void *) buffer;
+
+#if USE_TRAP
+	asm volatile (	"trap #15"
+					: "=r" (d0), "=r" (d1)
+					: "r" (d0), "r" (d1), "r" (d2), "r" (a0), "r" (a1)
+					: "d0", "d1", "d2", "a0", "a1", "memory"
+	);
+#else
+	asm volatile (	"jsr (%2)"
+					: "=r" (d0), "=r" (d1)
+					: "a" (BVMBugEntry), "r" (d0), "r" (d1), "r" (d2), "r" (a0), "r" (a1)
+					: "d0", "d1", "d2", "a0", "a1", "memory"
+	);
+#endif
+
+	*count = d1;
+
+	return d0;
+}
+
+/*--------------------------------------------------------------------------*/
+
+static
+__inline__
+const NETBOOTINFO *
+BVMBug_netbootinfo
+(	void
+)
+{	register unsigned long	d0 asm("d0") = (unsigned long) BB_NETBOOTINFO;
+	register NETBOOTINFO *	a0 asm("a0");
+
+#if USE_TRAP
+	asm volatile (	"trap #15"
+					: "=r" (a0)
+					: "r" (d0)
+					: "d0", "a0"
+	);
+#else
+	asm volatile (	"jsr (%1)"
+					: "=r" (a0)
+					: "a" (BVMBugEntry), "r" (d0)
+					: "d0", "a0"
+	);
+#endif
+
+	return a0;
+}
+
+/*--------------------------------------------------------------------------*/
+
+static
+__inline__
+const char *
+BVMBug_bootargs
+(	void
+)
+{	register unsigned long	d0 asm("d0") = (unsigned long) BB_BOOTARGS;
+	register const char *	a0 asm("a0");
+
+#if USE_TRAP
+	asm volatile (	"trap #15"
+					: "=r" (a0)
+					: "r" (d0)
+					: "d0", "a0"
+	);
+#else
+	asm volatile (	"jsr (%1)"
+					: "=r" (a0)
+					: "a" (BVMBugEntry), "r" (d0)
+					: "d0", "a0"
+	);
+#endif
+
+	return a0;
+}
+
 /*-----------------------------< end of file >------------------------------*/
