@@ -1110,6 +1110,49 @@ case_equal_strings
 	return TRUE;
 }
 
+/*--------------------------------------------------------------------------*/
+/* Replace '{}' sequence in source with bootrom arguments in destination
+ */
+
+int
+expand_bootrom_args
+(	char		*dest,
+	const char	*srce,
+	int			maxlen
+)
+{
+	/* leave space for NULL terminator */
+	maxlen--;
+
+	while (*srce && maxlen)
+	{
+		if (srce[0] == '{' && srce[1] == '}')
+		{
+			const char *p = bootrom_args();
+
+			while (maxlen && *p)
+			{
+				*dest++ = *p++;
+				maxlen--;
+			}
+
+			if (*p)
+				break;
+
+			srce += 2;
+		}
+		else
+		{
+			*dest++ = *srce++;
+			maxlen--;
+		}
+	}
+
+	*dest = '\0';
+
+	return *srce ? -1 : 0;
+}
+
 /*--------------------------------------------------------------------------*/
 
 int
@@ -1452,7 +1495,7 @@ file_close
 /*--------------------------------------------------------------------------*/
 
 void
-percent_init
+progress_init
 (	unsigned long	total
 )
 {
@@ -1464,7 +1507,7 @@ percent_init
 /*--------------------------------------------------------------------------*/
 
 void
-percent_term
+progress_term
 (	const char	*msg
 )
 {
@@ -1481,7 +1524,7 @@ percent_term
 /*--------------------------------------------------------------------------*/
 
 void
-percent_show
+progress_show
 (	unsigned long	current
 )
 {	unsigned 		percentage;
